/* *INDENT-OFF* */
!/* PSPP - computes sample statistics.
!   Copyright (C) 1997 Free Software Foundation, Inc.
!
!   This program is free software; you can redistribute it and/or
!   modify it under the terms of the GNU General Public License as
!   published by the Free Software Foundation; either version 2 of the
!   License, or (at your option) any later version.
!
!   This program is distributed in the hope that it will be useful, but
!   WITHOUT ANY WARRANTY; without even the implied warranty of
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
!   General Public License for more details.
!
!   You should have received a copy of the GNU General Public License
!   along with this program; if not, write to the Free Software
!   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!   02111-1307, USA.
!
!   The author may be contacted at <pfaffben@pilot.msu.edu> on the
!   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
!   through more mundane means. */
!
$ crosstabs:
$   *tables=custom;
$   +variables=custom;
$   +missing=miss:!table/include/report;
$   +write[wr_]=none,cells,all;
$   +format=fmt:!labels/nolabels/novallabs,
$           val:!avalue/dvalue,
$           indx:!noindex/index,
$           tabl:!tables/notables,
$           box:!box/nobox;
$   +cells[cl_]=count,none,row,column,total,expected,resid,sresid,asresid,all;
$   +pivot=pivot:!on/off;
$   +statistics[st_]=chisq,phi,cc,lambda,uc,none,btau,ctau,risk,gamma,d,
$                    kappa,eta,corr,all.
/* *INDENT-ON* */

/*#undef DEBUGGING*/
/*#define DEBUGGING 1 */
#include "debug-print.h"

/* Main declarations. */

#include <config.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include "avl.h"
#include "arena.h"
#include "hash.h"
#include "var.h"
#include "vfm.h"
#include "arena.h"
/* (headers) */
/* (decls global) */

/* A single table entry for general mode. */
typedef struct table_entry
  {
    int table;		/* Flattened table number. */
    double freq;	/* Frequency count. */
    value v[1];		/* Values. */
  }
table_entry;

/* A crosstabulation. */
typedef struct crosstab
  {
    int nvar;		/* Number of variables. */
    variable *v[2];	/* At least two variables; sorted by
			   larger indices first. */
  }
crosstab;

/* Indexes into crosstab.v. */
enum
  {
    ROW_VAR = 0,
    COL_VAR = 1,
  };

/* General mode crosstabulation table. */
hsh_table *gen_tab;			/* Hash table. */
table_entry **sorted_tab;		/* Sorted table. */

/* Integer mode crosstabulation table. */
double ***int_tab;

/* VARIABLES dictionary. */
static dictionary *var_dict;

/* TABLES. */
static crosstab **xtab;
static int nxtab;

/* Integer or general mode? */
enum
  {
    INTEGER,
    GENERAL,
  };
static int mode;

/* CELLS. */
static int no_cells;		/* No cells are requested. */

/* WRITE. */
static int write;		/* One of WR_* that specifies the WRITE style. */

/* Arenas. */
static arena *ar_tc;		/* For table cells. */
static arena *ar_col;		/* For column data. */

static int internal_cmd_crosstabs (void);
static void free_var_dict (void);
static void precalc (void);
static int calc_general (ccase *);
static int calc_integer (ccase *);
static void postcalc (void);

#if DEBUGGING
static void debug_print (void);
static void print_table_entries (table_entry **tab);
#endif

/* Parse and execute CROSSTABS, then clean up. */
int
cmd_crosstabs (void)
{
  int result = internal_cmd_crosstabs ();

  free_var_dict ();
  arena_destroy (&ar_tc);
  arena_destroy (&ar_col);
  
  return result;
}

/* Parses and executes the CROSSTABS procedure. */
static int
internal_cmd_crosstabs (void)
{
  var_dict = NULL;
  xtab = NULL;
  nxtab = 0;
  arena_create (&ar_tc);
  arena_create (&ar_col);
  /* (parser) */

#if DEBUGGING
  /* Needs var_dict. */
  debug_print ();
#endif

  mode = var_dict ? INTEGER : GENERAL;
  free_var_dict();

  /* CELLS. */
  no_cells = 0;
  if (!sbc_cells)
    cl_a_cells[CL_COUNT] = 1;
  else 
    {
      int i;
      int count = 0;

      for (i = 0; i < CL_count; i++)
	if (cl_a_cells[i])
	  count++;
      if (count == 0)
	{
	  cl_a_cells[CL_COUNT] = 1;
	  cl_a_cells[CL_ROW] = 1;
	  cl_a_cells[CL_COLUMN] = 1;
	  cl_a_cells[CL_TOTAL] = 1;
	}
      if (cl_a_cells[CL_ALL])
	for (i = 0; i < CL_count; i++)
	  cl_a_cells[i] = 1;
      if (count == 1 && cl_a_cells[CL_NONE])
	no_cells = 1;
    }

  /* STATISTICS. */
  if (sbc_statistics)
    {
      int i;
      int count = 0;

      for (i = 0; i < ST_count; i++)
	if (st_a_statistics[i])
	  count++;
      if (count == 0)
	st_a_statistics[ST_CHISQ] = 1;
      if (st_a_statistics[ST_ALL])
	for (i = 0; i < ST_count; i++)
	  st_a_statistics[i] = 1;
    }
  
  /* MISSING. */
  if (miss == REPORT && mode == GENERAL)
    {
      msg (SE, _("Missing mode REPORT not allowed in general mode.  "
		 "Assuming MISSING=TABLE."));
      miss = TABLE;
    }

  /* WRITE. */
  if (wr_a_write[WR_ALL] && wr_a_write[WR_CELLS])
    wr_a_write[WR_ALL] = 0;
  if (wr_a_write[WR_ALL] && mode == GENERAL)
    {
      msg (SE, _("Write mode ALL not allowed in general mode.  "
		 "Assuming WRITE=CELLS."));
      wr_a_write[WR_CELLS] = 1;
    }
  if (sbc_write
      && wr_a_write[WR_NONE] + wr_a_write[WR_ALL] + wr_a_write[WR_CELLS] == 0)
    wr_a_write[WR_CELLS] = 1;
  if (wr_a_write[WR_CELLS])
    write = WR_CELLS;
  else if (wr_a_write[WR_ALL])
    write = WR_ALL;
  else
    write = WR_NONE;

  update_weighting (&default_dict);
  procedure (precalc, mode == GENERAL ? calc_general : calc_integer, postcalc);

  return 1;
}

/* Frees var_dict once it's no longer needed. */
static void
free_var_dict (void)
{
  if (!var_dict)
    return;
  
  {
    int i;

    if (var_dict->var_by_name)
      {
	avl_destroy (var_dict->var_by_name, NULL);
	var_dict->var_by_name = NULL;
      }

    for (i = 0; i < var_dict->nvar; i++)
      free (var_dict->var[i]);
    free (var_dict->var);
    var_dict->var = NULL;
    var_dict->nvar = 0;

    free_dictionary (var_dict);

    var_dict = NULL;
  }
}

/* Parses the TABLES subcommand. */
static int
custom_tables (void)
{
  dictionary *dict;
  int n_by;
  variable ***by = NULL;
  int *by_nvar = NULL;
  int nx = 1;
  int success = 0;

  /* Ensure that this is a TABLES subcommand. */
  if (!match_id (TABLES) && (token != ID || !is_varname (tokstr))
      && token != ALL)
    return 2;
  match_tok ('=');

  dict = var_dict ? var_dict : &default_dict;
  
  for (n_by = 0; ;)
    {
      by = xrealloc (by, sizeof (variable **) * (n_by + 1));
      by_nvar = xrealloc (by_nvar, sizeof (int) * (n_by + 1));
      if (!parse_variables (dict, &by[n_by], &by_nvar[n_by],
			    PV_NO_DUPLICATE | PV_NO_SCRATCH))
	goto lossage;
      nx *= by_nvar[n_by];
      n_by++;

      if (!match_tok (BY))
	{
	  if (n_by < 1)
	    {
	      syntax_error (_("expecting BY"));
	      goto lossage;
	    }
	  else 
	    break;
	}
    }
  
  {
    int *by_iter = xcalloc (sizeof (int) * n_by);
    int i;

    xtab = xrealloc (xtab, sizeof (variable *) * (nxtab + nx));
    for (i = 0; i < nx; i++)
      {
	crosstab *x = xmalloc (sizeof (crosstab)
			       + sizeof (variable *) * (n_by - 2));
	x->nvar = n_by;
	
	{
	  int i;
	  
	  for (i = 0; i < n_by; i++)
	    x->v[i] = by[i][by_iter[i]];
	}
	
	{
	  int i;

	  for (i = n_by - 1; i >= 0; i--)
	    {
	      if (++by_iter[i] < by_nvar[i])
		break;
	      by_iter[i] = 0;
	    }
	}

	xtab[nxtab++] = x;
      }
    free (by_iter);
  }
    
  success = 1;
  /* Despite the name, we come here whether we're successful or
     not. */
 lossage:
  {
    int i;

    for (i = 0; i < n_by; i++)
      free (by[i]);
    free (by);
    free (by_nvar);
  }

  return success;
}

/* Parses the VARIABLES subcommand. */
static int
custom_variables (void)
{
  variable **v = NULL;
  int nv = 0;

  if (nxtab)
    return msg (SE, _("VARIABLES must be specified before TABLES."));

  match_tok ('=');
  
  for (;;)
    {
      int orig_nv = nv;
      int i;

      long min, max;
      
      if (!parse_variables (&default_dict, &v, &nv,
			    (PV_APPEND | PV_NUMERIC
			     | PV_NO_DUPLICATE | PV_NO_SCRATCH)))
	return 0;

      if (token != '(')
	{
	  syntax_error ("expecting `('");
	  goto lossage;
	}
      get_token ();

      if (token != NUM || tokint == NOT_LONG)
	{
	  syntax_error ("expecting integer");
	  goto lossage;
	}
      min = tokint;
      get_token ();

      match_tok (',');

      if (token != NUM || tokint == NOT_LONG)
	{
	  syntax_error ("expecting integer");
	  goto lossage;
	}
      max = tokint;
      if (max < min)
	{
	  msg (SE, _("Maximum value (%ld) less than minimum value (%ld)."),
	       max, min);
	  goto lossage;
	}
      get_token ();

      if (token != ')')
	{
	  syntax_error ("expecting `)'");
	  goto lossage;
	}
      get_token ();
      
      for (i = orig_nv; i < nv; i++)
	{
	  v[i]->p.crs.min = min;
	  v[i]->p.crs.max = max;
	}
      
      if (token == '/')
	break;
    }
  
  {
    int i;
    
    var_dict = new_dictionary (0);
    var_dict->var = xmalloc (sizeof (variable *) * nv);
    var_dict->nvar = nv;
    for (i = 0; i < nv; i++)
      {
	variable *var = xmalloc (offsetof (variable, width));
	strcpy (var->name, v[i]->name);
	var->index = i;
	var->type = v[i]->type;
	var->foo = v[i]->index;
	var_dict->var[i] = var;
	force_avl_insert (var_dict->var_by_name, var);
      }

    free (v);
    return 1;
  }

 lossage:
  free (v);
  return 0;
}

#if DEBUGGING
static void
debug_print (void)
{
  printf ("CROSSTABS\n");

  if (var_dict)
    {
      int i;

      printf ("\t/VARIABLES=");
      for (i = 0; i < var_dict->nvar; i++)
	{
	  variable *v = var_dict->var[i];
	  variable *iv = default_dict.var[v->foo];

	  printf ("%s ", v->name);
	  if (i < var_dict->nvar - 1)
	    {
	      variable *nv = var_dict->var[i + 1];
	      variable *niv = default_dict.var[nv->foo];
	      
	      if (iv->p.crs.min == niv->p.crs.min
		  && iv->p.crs.max == niv->p.crs.max)
		continue;
	    }
	  printf ("(%ld,%ld) ", (long) iv->p.crs.min, (long) iv->p.crs.max);
	}
      printf ("\n");
    }
  
  {
    int i;

    printf ("\t/TABLES=");
    for (i = 0; i < nxtab; i++)
      {
	crosstab *x = xtab[i];
	int j;

	if (i)
	  printf("\t\t");
	for (j = 0; j < x->nvar; j++)
	  {
	    if (j)
	      printf (" BY ");
	    printf ("%s", x->v[j]->name);
	  }
	printf ("\n");
      }
  }
}
#endif /* DEBUGGING */

/* Data file processing. */

static int compare_table_entry (const void *, const void *, void *);
static unsigned hash_table_entry (const void *, void *);

/* Set up the crosstabulation tables for processing. */
static void
precalc (void)
{
  if (mode == GENERAL)
    {
      gen_tab = hsh_create (512, compare_table_entry, hash_table_entry,
			    NULL, NULL);
    }
  else 
    {
      /* FIXME: still wasn't right. */
    }
}

/* Form crosstabulations for general mode. */
static int
calc_general (ccase *c)
{
  /* Case weight. */
  double w = (default_dict.weight_index != -1
	      ? c->data[default_dict.var[default_dict.weight_index]->fv].f
	      : 1.0);

  /* Flattened current table index. */
  int t;

  for (t = 0; t < nxtab; t++)
    {
      const crosstab *x = xtab[t];
      const size_t entry_size = (sizeof (table_entry)
				 + sizeof (value) * (x->nvar - 1));
      table_entry *te = local_alloc (entry_size);

      /* Construct table entry for the current record and table. */
      te->table = t;
      {
	int j;

	assert (x);
	for (j = 0; j < x->nvar; j++)
	  if (x->v[j]->type == NUMERIC)
	    te->v[j].f = c->data[x->v[j]->fv].f;
	  else
	    {
	      memcpy (te->v[j].s, c->data[x->v[j]->fv].s, x->v[j]->width);
	      
	      /* Necessary because of the really dumb hash routine. */
	      memset (&te->v[j].s[x->v[j]->width], 0,
		      sizeof (value) - x->v[j]->width);
	    }
      }

      /* Add record to hash table. */
      {
	table_entry **tepp = (table_entry **) hsh_probe (gen_tab, te);
	if (NULL == *tepp)
	  {
	    table_entry *tep = arena_alloc (&ar_tc, entry_size);
	    
	    te->freq = w;
	    memcpy (tep, te, entry_size);
	    
	    *tepp = tep;
	  }
	else
	  (*tepp)->freq += w;
      }
      local_free (te);
    }
  
  return 1;
}

static int
calc_integer (ccase *c)
{
  return 1;
}

#if DEBUGGING
/* Print out all table entries in NULL-terminated TAB for use by a
   debugger (a person, not a program). */
static void
print_table_entries (table_entry **tab)
{
  printf ("raw crosstabulation data:\n");
  for (; *tab; tab++)
    {
      const crosstab *x = xtab[(*tab)->table];
      int i;

      printf ("(%g) table:%d ", (*tab)->freq, (*tab)->table);
      for (i = 0; i < x->nvar; i++)
	{
	  if (i)
	    printf (", ");
	  printf ("%s:", x->v[i]->name);
	  
	  if (x->v[i]->type == NUMERIC)
	    printf ("%g", (*tab)->v[i].f);
	  else
	    printf ("%.*s", x->v[i]->width, (*tab)->v[i].s);
	}
      printf ("\n");
    }
}
#endif

/* Compare the table_entry's at _A and _B and return a strcmp()-type
   result. */
static int 
compare_table_entry (const void *_a, const void *_b, void *foo)
{
  const table_entry *a = (const table_entry *) _a;
  const table_entry *b = (const table_entry *) _b;
  
  {
    const int difftable = a->table - b->table;
    if (difftable)
      return difftable;
  }
  
  {
    const crosstab *x = xtab[a->table];
    int i;

    static int comparing = 0;
    for (i = x->nvar - 1; i >= 0; i--)
      if (x->v[i]->type == NUMERIC)
	{
	  const double diffnum = a->v[i].f - b->v[i].f;
	  if (diffnum < 0)
	    return -1;
	  else if (diffnum > 0)
	    return 1;
	}
      else 
	{
	  assert (x->v[i]->type == ALPHA);
	  {
	    const int diffstr = strncmp (a->v[i].s, b->v[i].s, x->v[i]->width);
	    if (diffstr)
	      return diffstr;
	  }
	}
  }
  
  return 0;
}

/* Calculate a hash value from table_entry _A. */
static unsigned
hash_table_entry (const void *_a, void *foo)
{
  const table_entry *a = (const table_entry *) _a;
  unsigned long hash = a->table;
  int i;

  /* Hash algorithm from _SPSS Statistical Algorithms_. */
  for (i = 0; i < xtab[a->table]->nvar; i++)
    {
      hash = (hash << 3) | (hash >> (CHAR_BIT * SIZEOF_LONG - 3));
      hash ^= a->v[i].hash[0];
#if SIZEOF_DOUBLE / SIZEOF_LONG > 1
      hash ^= a->v[i].hash[1];
#endif
    }
  
  return hash;
}

/* Post-data reading calculations. */

static int find_pivot_extent (table_entry ***beg, int *cnt);
static void enum_column_values (table_entry **beg, int cnt,
				value **values, int *nvalues);

static void
postcalc (void)
{
  if (mode == GENERAL)
    {
      sorted_tab = (table_entry **) hsh_sort (gen_tab, compare_table_entry);
#if DEBUGGING
      print_table_entries (sorted_tab);
#endif
    }
  else
    {
      assert (0);
    }
  
  /* Identify all the individual crosstabulation tables, and deal with
     them. */
  {
    table_entry **beg = NULL;
    int cnt = 0;

    while (find_pivot_extent (&beg, &cnt))
      {
	value *cols;
	int ncols;
	double *mat;
	
	enum_column_values (beg, cnt, &cols, &ncols);
	printf ("beg=%d, cnt=%d\n", beg - sorted_tab, cnt);

#if DEBUGGING
	{
	  int i;

	  printf ("column values:");
	  for (i = 0; i < ncols; i++)
	    printf (" %g", cols[i].f);
	  printf ("\n");
	}
#endif
	
	/*mat = xmalloc (sizeof (double) * (a->n + 1) * (b->n + 1));*/

	free (cols);
      }
  }
  
  hsh_destroy (gen_tab);
}

/* Find the extent of a region in sorted_tab that contains one table.
   If we're outputting pivot tables then that means a set of table
   entries with identical table number; otherwise they also have to
   have the same values for every dimension after the row and column
   dimensions.  *BEG receives a pointer to the first table entry, and
   *CNT receives the number of entries to be grouped together.  At
   first call BEG should be NULL, *cnt should be 0.  Returns 0 if
   sorted_tab is exhausted. */
static int
find_pivot_extent (table_entry ***beg, int *cnt)
{
  if (*beg == NULL)
    *beg = sorted_tab;
  else
    *beg = &(*beg)[*cnt];
  if ((*beg)[0] == NULL)
    return 0;
  
  for (*cnt = 1; ; (*cnt)++)
    {
      if ((*beg)[*cnt] == NULL || (*beg)[*cnt]->table != (**beg)->table)
	return 1;
      if (pivot == ON)
	continue;
      
      {
	crosstab *x = xtab[(*beg)[*cnt]->table];
	int i;
	  
	for (i = 0; i < x->nvar; i++)
	  if (x->v[i]->type == NUMERIC)
	    {
	      if ((*beg)[*cnt]->v[i].f != (**beg)->v[i].f)
		return 1;
	    }
	  else 
	    {
	      if (strncmp ((*beg)[*cnt]->v[i].s, (**beg)->v[i].s,
			   x->v[i]->width))
		return 1;
	    }
      }
    }
}

/* Compare value * A and B, where WIDTH is the string width or 0 for
   numerics, and return a strcmp()-type result. */
static int
compare_value (const void *_a, const void *_b, void *_width)
{
  const value *a = _a;
  const value *b = _b;
  const int width = (int) _width;

  if (width)
    return strncmp (a->s, b->s, width);
  else
    {
      const double diffnum = a->f - b->f;

      if (diffnum < 0)
	return -1;
      else if (diffnum > 0)
	return 1;
      else
	return 0;
    }
}

/* Given a list of CNT table_entry's starting at BEG, creates a list
   of *NVALUES column values *VALUES. */
static void 
enum_column_values (table_entry **beg, int cnt, value **values, int *nvalues)
{
  avl_tree *tree = avl_create (&ar_col, compare_value,
			       ((void *)
				(xtab[(*beg)->table]->v[COL_VAR]->width)));

  int i;
  
#if 0
  for (i = 0; i < cnt; i++)
    avl_insert (tree, 
#endif
}

/* Local Variables: */
/* mode:c */
/* End: */
